/*******************************************************************************

        @file NodeParser.d
        
        Copyright (c) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.

        4. Derivative works are permitted, but they must carry this notice
           in full and credit the original source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version; March 2005      

        @author         Kris


*******************************************************************************/

module example.BlockParser;

/*******************************************************************************

        Parses a text string into a set of related text blocks. This can
        be used in conjunction with code-processing tools.

*******************************************************************************/

class Parser 
{
        private Node root;

        private typedef Node delegate(Node parent, Node prev) Allocator;

        private enum State {Begin, SQuote, DQuote, Comment, Comment1, Comment2, Comment3};

        /***********************************************************************

                Structure to represent a block of text, along with the 
                relationship between this block and others around it.

        ***********************************************************************/

        class Node 
        {
                // 'statement' is the default!
                enum Type {Statement, Function, Literal, Class, Comment, Root};

                Node    prev,
                        next,
                        child,
                        parent;

                Object  info;
                Type    type;
                char[]  label, 
                        content,
                        context;
                char[]  bracket;

                /***************************************************************

                        Hook up a block to its parent and its prior
                        sibling, if any.

                ***************************************************************/

                this (Node parent, Node prev)
                {
                        bracket = "{}";

                        // hook up parent
                        if ((this.parent = parent) !== null)
                             if (! parent.child)
                                   parent.child = this;

                        // hook up sibling
                        if ((this.prev = prev) !== null)
                           {
                           this.next = prev.next;
                           prev.next = this;
                           }
                }

                /***************************************************************

                        Hook up a block to its parent 

                ***************************************************************/

                Node addChild ()
                {       
                        Node prev = child ? child.lastChild : null;
                        return new Node (this, prev);     
                }

                /***************************************************************

                        Hook up a block to its parent 

                ***************************************************************/

                void move (Node parent)
                {
                        if (parent.child)
                           {
                           Node node = lastChild;
                           (node.next = parent.child).prev = this;
                           }
                        parent.child = this;
                }

                /***************************************************************

                        Find last sibling

                ***************************************************************/

                Node lastChild ()
                {
                        Node tmp = this;

                        while (tmp.next)
                               tmp = tmp.next;
                        return tmp;
                }

                /***************************************************************

                        visit each node of the tree, starting with the 
                        provided block.

                ***************************************************************/

                void visit (void delegate(Node) dg)
                {
                        visit (this, dg);
                }

                /***************************************************************
        
                        Tree traversal routine

                ***************************************************************/

                private static final void visit (Node block, void delegate(Node) dg)
                {
                        while (block)
                              {
                              if (block.type != block.Type.Comment)
                                  dg (block);

                              if (block.child)
                                  visit (block.child, dg);
                              block = block.next;
                              }
                }
        }

        /***********************************************************************

        ***********************************************************************/

        Node parse (void[] content)
        {
                Node host = allocator (null, null);
                host.type = Node.Type.Root;

                parse (host, null, cast(char[]) content);
                return (root = host.child);
        }

        /***********************************************************************

        ***********************************************************************/

        Node getRoot ()
        {
                return root;
        }

        /***********************************************************************

        ***********************************************************************/

        private static Node allocator (Node parent, Node child)
        {
                return new Node (parent, child);
        }
         
        /***********************************************************************

                Parse the given content into a set of related text blocks.
                Each block may have a parent, a preceding and/or following
                sibling, and an optional child. The child may have similar
                properties.

                These blocks are split based upon the content held therein.
                For example, /* comments are placed into distinct blocks
                from compound statements. The latter is denoted by brace
                characters ('{' '}'), but could be replaced with others.

                The idea here is to split the text into a semi-structured 
                representation for further processing.

        ***********************************************************************/

        private int parse (Node parent, Node prior, char[] content)
        {
                int       mark,
                          done, 
                          indent;
                char      lastChar;
                char[]    keyword;
                Node      current;

                State state = State.Begin;
                current = allocator (parent, prior);
                for (int i; i < content.length; ++i)
                    {
/+
                    void pushComment ()
                    {
                            current.content = content[done..i-1];
                            printf ("precomment [%.*s]", current.content);
                            current = allocator (parent, current);
                            done = i - 1;
                    }

                    void popComment (int extra)
                    {
                            current.content = content[done..i+extra];
                            current.type = Node.Type.Comment;
                            printf ("postcomment [%.*s]", current.content);
                            current = allocator (parent, current);
                            done = i + extra;
                    }
+/
                    char c = content.ptr[i];
                    switch (state)
                           {
                           case State.Begin:                           
                                if (c == '"')
                                    state = State.DQuote;
                                else
                                if (c == '\'')
                                    state = State.SQuote;
                                else
                                if (c == '/')
                                    state = State.Comment;
                                else
                                if (c == '{')
                                   {
                                   current.content = content[done..i];
                                   //printf ("recursing [%.*s]", current.content);
                                   i = i + 1 + parse (current, null, content[i+1..length]);
                                   current = allocator (parent, current);
                                   done = i+1;
                                   }
                                else
                                if (c == '}')
                                   {
                                   current.content = content[done..i];
                                   //printf ("returning [%.*s]", current.content);
                                   return i;
                                   }
                                break;

                           case State.SQuote:
                                if (c == '\'' && lastChar != '\\')
                                    state = State.Begin;
                                break;

                           case State.DQuote:
                                if (c == '"' && lastChar != '\\')
                                    state = State.Begin;
                                break;

                           case State.Comment:
                                if (c == '*')
                                   {
                                   state = State.Comment1;
                                   current.content = content[done..i-1];
                                   //printf ("precomment [%.*s]", current.content);
                                   current = allocator (parent, current);
                                   done = i - 1;
                                   }
                                else
                                   if (c == '/')
                                      {
                                      state = State.Comment3;
                                      //pushComment();
                                      }
                                   else
                                      state = State.Begin;
                                break;

                           case State.Comment1:                                                           
                                if (c == '*') 
                                    state = State.Comment2;
                                break;

                           case State.Comment2:
                                if (c == '/') 
                                   {
                                   state = State.Begin;
                                   current.type = Node.Type.Comment;
                                   current.content = content[done .. i + 1];
                                   //printf ("postcomment [%.*s]", current.content);
                                   current = allocator (parent, current);
                                   done = i + 1;
                                   }
                                else
                                   state = State.Comment1;
                                break;

                           case State.Comment3:
                                if (c == '\n' || c == '\r') 
                                   {                                                                  
                                   state = State.Begin; 
                                   //popComment (0);
                                   }
                                break;
                           }
                      lastChar = c;
                      }
                return content.length;
        }
}

