module example.Processor;

private import  std.ctype;
private import  std.regexp;

private import  example.BlockParser;

private import  mango.utils.Text;

private import  mango.format.Int;

private import  mango.io.Stdout,
                mango.io.File,
                mango.io.FileScan,
                mango.io.FileSystem,
                mango.io.FileConduit;

version = FileOutput;

/*******************************************************************************

        This is prefixed appropriately to import statements

*******************************************************************************/

const char[] IMPORT_PREFIX = "mango";

/*******************************************************************************

        struct for text replacements

*******************************************************************************/

struct SimplePatch
{
        char[]          from,
                        into;
}

/*******************************************************************************

        Superclass for more complex replacements

*******************************************************************************/

class CustomPatch : Replace
{
        this (char[] pattern) {super(pattern);}

        abstract char[] convert (char[] content);
}


/*******************************************************************************

*******************************************************************************/

class Processor 
{    
        alias void delegate (Parser.Node) Delegate;

        private FilePath                path;
        private char[]                  theFile,
                                        filename;
        private bool                    multiFile;
        private bool[char[]]            Keywords;
        private FileRegistry[char[]]    fileRegistry;
        private ParsedFile[]            parsedFiles;
        private ParsedFile              currentParser;
        private PackageMember[char[]]   packageMembers;

        private int                     ctors,
                                        files,
                                        lines,
                                        functions,
                                        ctorsAdded,
                                        importsAdded,
                                        staticCtorsAdded;

        private FuncDecl                funcDecl;
        private ClassDecl               classDecl;
        private LiteralDecl             literalDecl;
        private InterfaceDecl           interfaceDecl;
        private StaticCtorDecl          staticCtorDecl;
        private StaticCtorRemove        staticCtorRemove;
        private EqualsNull              equalsNull;
        private Extends                 extends;
        private Throws                  throws;
        private String                  string;
        private Import                  imports;
        private Package                 moduleName;
        private PackagePass2            packagePass2;
        private StaticIntArray          staticIntArray;
        private ArrayCopy               arrayCopy;
        private StaticInitializer       staticInitializer;
        private InstanceInitializer     instanceInitializer;
        private NewArrayInstance        newArrayInstance;
        private LiteralStringEquals     literalStringEquals;
        private RenameBoolean           renameBoolean;
        private RenameKeyword           renameKeyword;           
        private RenameError             renameError;           
        private RenameChar              renameChar;           
        private RemoveFinal             removeFinal;                      
        private ReturnCast              returnCast;
        private UberCast                uberCast;
        private Instanceof              instanceof;
        private ToSwtString             toSwtString;
        private ImportReplace           importReplace;

        /***********************************************************************

                Additional info attached to class nodes

        ***********************************************************************/

        class ClassAttrib
        {
                Parser.Node[]  ctors;
        }

        /***********************************************************************

                These are created for pass2

        ***********************************************************************/

        class ParsedFile
        {       
                int             errors;
                FilePath        path;
                Parser          parser;
                char[]          source;
                char[][]        classes;
                char[]          moduleName;
                PackageMember   packageMember;

                this (FilePath path, char[] source, Parser parser)
                {
                        this.path = path;
                        this.source = source;
                        this.parser = parser;
                }

                void addClass (char[] name)
                {
                        classes ~= name;
                }

                void setModule (char[] moduleName, PackageMember packageMember)
                {
                        this.moduleName = moduleName;
                        this.packageMember = packageMember;
                }

                void message (char[] msg)
                {
                        if (++errors == 1)
                            Stdout (path.toString) (CR);
                        Stdout ("  ") (msg) (CR);
                }
        }

        /***********************************************************************

                These are created for pass2

        ***********************************************************************/

        class PackageMember
        {
                ParsedFile[]    files;
                char[]          moduleName;
                char[]          packageName;

                this (char[] packageName)
                {
                        this.packageName = packageName;
                }

                void addModule (ParsedFile file, char[] moduleName)
                {
                        files ~= file;
                        file.setModule (moduleName, this);
                }
        }

        /***********************************************************************

        ***********************************************************************/

        interface Transform
        {
                abstract char[] invoke (char[] source);
        }

        /***********************************************************************

        ***********************************************************************/

        class SimpleTransform : Replace, Transform
        {
                private SimplePatch patch;
                 
                this (SimplePatch patch)
                {
                        this.patch = patch;
                }

                char[] invoke (char[] source)
                {
                        int i, j;                        
                        j = count;

                        set (source);
                        while ((i = Text.indexOf(content, patch.from)) >= 0)
                              {
                              ++count;
                              sections ~= content [0 .. i];
                              sections ~= patch.into;
                              content = content [i+patch.from.length .. length];
                              }
                        if (count == j)
                            printf ("  failed to locate '%.*s'\n", patch.from);

                        return combine;
                }
        }

        /***********************************************************************

        ***********************************************************************/

        class CustomTransform : Transform
        {
                private CustomPatch patch;
                 
                this (CustomPatch patch)
                {
                        this.patch = patch;
                }

                char[] invoke (char[] source)
                {
                        return patch.convert (source);
                }
        }

        /***********************************************************************

        ***********************************************************************/

        class IgnoreTransform : Transform
        {
                char[] invoke (char[] source)
                {
                        return null;
                }
        }

        /***********************************************************************

        ***********************************************************************/

        class FileRegistry
        {
                FileRegistry    next;
                Transform       transform;
                
                this (FileRegistry next, Transform transform)
                {
                        this.next = next;
                        this.transform = transform;
                }

                char[] invoke (char[] source)
                {
                        FileRegistry list = this;

                        while (list)
                              {
                              source = list.transform.invoke (source);
                              if (source is null)
                                  return null;
                              list = list.next;
                              }
                        return source;
                }
        }

        /***********************************************************************

        ***********************************************************************/

        this (FilePath path)
        {
                this.path = path;

                Keywords["if"] = true;
                Keywords["while"] = true;
                Keywords["for"] = true;
                Keywords["catch"] = true;
                Keywords["switch"] = true;

                funcDecl = new FuncDecl;
                classDecl = new ClassDecl (this);
                literalDecl = new LiteralDecl;
                interfaceDecl = new InterfaceDecl;
                staticCtorDecl = new StaticCtorDecl;
                staticCtorRemove = new StaticCtorRemove;

                throws = new Throws;
                string = new String;
                extends = new Extends;
                imports = new Import (this);
                arrayCopy = new ArrayCopy;
                equalsNull = new EqualsNull;
                moduleName = new Package (this);
                packagePass2 = new PackagePass2 (this);
                newArrayInstance = new NewArrayInstance;
                staticIntArray = new StaticIntArray;
                staticInitializer = new StaticInitializer;
                instanceInitializer = new InstanceInitializer;
                literalStringEquals = new LiteralStringEquals;
                renameBoolean = new RenameBoolean;
                renameKeyword = new RenameKeyword;
                renameChar = new RenameChar;
                removeFinal = new RemoveFinal;
                renameError = new RenameError;
                uberCast = new UberCast;
                returnCast = new ReturnCast;
                instanceof = new Instanceof;
                toSwtString = new ToSwtString;
                importReplace = new ImportReplace;
        }

        /***********************************************************************

        ***********************************************************************/

        public void start ()
        {
                if (path.getExtension.length)
                    pass1 (new File (path));
                else
                   {
                   multiFile = true;
                   FileScan scan = new FileScan;
                   scan (path, "java");
                   scan.directories (&directory);
                   scan.files (&pass1);
                   }

                pass2 ();

                int count;
                char[] stats = statistics(count);
                printf ("%.*s-----\n%d\tchanges made\n", stats, count);
        }

        /***********************************************************************

                Patch a filename with respect to the processing path

        ***********************************************************************/

        protected char[] normalizeFile (char[] file)
        {
                file = FilePath.normalize (file);
                file = (new FilePath(file)).splice(path);
                return file;
        }

        /***********************************************************************

                Register a file for customization

        ***********************************************************************/

        private void register (char[] file, Transform transform)
        {
                FileRegistry prior;

                file = normalizeFile (file.dup);
                if (file in fileRegistry)
                    prior = fileRegistry [file];
                fileRegistry[file] = new FileRegistry (prior, transform);
        }
        
        /***********************************************************************

                Register a file for text replacement

        ***********************************************************************/

        void replace (char[] file, CustomPatch patch)
        {
                register (file, new CustomTransform (patch));
        }
        
        /***********************************************************************

                Register a file for text replacement

        ***********************************************************************/

        void replace (char[] file, inout SimplePatch patch)
        {
                register (file, new SimpleTransform (patch));
        }
        
        /***********************************************************************

                Register a file for text replacement

        ***********************************************************************/

        void replace (char[] file, inout SimplePatch[] patches)
        {
                foreach (SimplePatch patch; patches)
                         replace (file, patch);
        }
        
        /***********************************************************************

                Register a file for text replacement

        ***********************************************************************/

        void ignore (char[] file)
        {
                register (file, new IgnoreTransform);
        }
        
        /***********************************************************************

                Convert input-file path to an output file

        ***********************************************************************/

        protected void nameOutputFile (MutableFilePath dst)
        {
                dst.setExt ("d");
        }

        /***********************************************************************

                Process a directory. This is for overriding.

        ***********************************************************************/

        protected void directory (FilePath path)
        {
        }

        /***********************************************************************

        ***********************************************************************/

        protected char[] preProcess (char[] source)
        {
                int i = 0;
                while ((i = Text.indexOf (source, '\n', i)) > 0)
                        ++i, ++lines;

                return source;
        }

        /***********************************************************************

                Process the provided file

        ***********************************************************************/

        protected void pass1 (File src)
        {
                ++files;
                theFile = src.getPath.toString;
                //Stdout (theFile) (CR);

                // convert to module name
                MutableFilePath mod = new MutableFilePath (src.getPath);
                mod.setExt(null).setRoot (null);

                filename = mod.toString;
                Text.replace (filename, FileSystem.PathSeperatorChar, '.');

                // construct output path
                MutableFilePath dst = new MutableFilePath (src.getPath);
                nameOutputFile (dst);

                // load the file content
                char[] source = preProcess (cast(char[]) src.read);

                // is this conversion customized?
                if (theFile in fileRegistry)
                   {
                   FileRegistry fr = fileRegistry [theFile];
                   printf ("custom processing for %.*s\n", theFile);
                   source = fr.invoke (source);

                   // skip this file ?
                   if (source is null)
                       return true;
                   }

                // parse into text blocks
                Parser parser = new Parser;
                Parser.Node root = parser.parse (source);

                // stash the parsed file for later reference
                currentParser = new ParsedFile (dst, source, parser);

                // push onto file list
                parsedFiles ~= currentParser;

                // label the code blocks
                root.visit (&label);
                
                // continue conversion
                root.visit (&preamble);
                root.visit (&constructor);
                root.visit (&general);
                root.visit (&decls);
                root.visit (&cleanup);
        }

        /***********************************************************************

        ***********************************************************************/

        void pass2()
        {
                ParsedFile file;

                void patchImports (Parser.Node node)
                {
                        if (node.parent.type == node.Type.Root)
                           {
                           node.content = packagePass2(imports(node.content, file), file);
                           }
                }

                //displayPackages();
                Stdout (CR) ("processing imports ...") ();
                foreach (ParsedFile pf; parsedFiles)
                        {
                        file = pf;
                        pf.parser.getRoot.visit (&patchImports);

                        output (pf.path, pf.parser.getRoot, multiFile);
//                        output (pf.path, pf.parser.getRoot, false);
                        }
                Stdout (CR) (CR);
        }

        /***********************************************************************

        ***********************************************************************/

        void displayPackages()
        {
                foreach (PackageMember pm; packageMembers)
                        {
                        Stdout (CR) (pm.moduleName) (':') (pm.packageName) (CR);
                        foreach (ParsedFile file; pm.files)
                                {
                                Stdout ("  ") (file.path) (CR);
                                foreach (char[] name; file.classes)
                                         Stdout ("    ") (name) (CR);
                                }
                        }
        }

        /***********************************************************************

                Scan the 'file' for references to class names within the
                package 'member', and construct an import statement that
                relects the required modules.

                This is used to replace import x.* idioms, which cause
                havoc when porting code. It also handles the case where
                Java implicitly imports all package classes, whilst the
                D language does not.

        ***********************************************************************/

        char[] buildImport (ParsedFile file, PackageMember member)
        {
                char[] imports = null;

                // TODO: how does this happen? 
                if (file.source is null)
                   {
                   file.message ("no source in " ~ file.path.toString);
                   return null;
                   }

                // see if any of these classes are present in the file. Add
                // specific import is so
                foreach (Processor.ParsedFile pf; member.files)
                         // don't match against ourself!
                         if (file.moduleName != pf.moduleName)
                             foreach (char[] className; pf.classes)
                                      if (Text.indexOf (file.source, className) >= 0)
                                         {
                                         ++importsAdded;
                                         imports = imports ~ "private import " ~ pf.moduleName ~ ";\n";
                                         break;
                                         }
                return imports;
        }

        /***********************************************************************

        ***********************************************************************/

        void output (FilePath dst, Parser.Node root, bool toFile)
        {
                // emit conversion
                if (toFile)       
                   {
                   FileConduit f = new FileConduit (dst, FileStyle.ReadWriteCreate);
                   IBuffer buffer = f.createBuffer();
                   emit (root, buffer);
                   buffer.flush();
                   f.close();
                   }
                else
                   {
                   emit (root, Stdout.getBuffer);
                   Stdout.flush();
                   }
        }

        /***********************************************************************

        ***********************************************************************/

        protected char[] getModuleName ()
        {
                return filename;
        }

        /***********************************************************************

                Construct statistics for this run

        ***********************************************************************/

        private char[] statistics (inout int count)
        {
                char[]  stats;

                count = ctors + ctorsAdded + importsAdded + staticCtorsAdded;

                stats ~= Replace.format (files, "files");
                stats ~= Replace.format (lines, "lines");
                stats ~= Replace.format (functions, "methods scanned");
                stats ~= Replace.format (ctors, "ctors patched");
                stats ~= Replace.format (ctorsAdded, "ctors added");
                stats ~= Replace.format (staticCtorsAdded, "static ctors added");
                stats ~= Replace.format (importsAdded, "explicit imports added");

                foreach (Replace r; Replace.handlers)
                        {
                        char[] tmp = r.stats();
                        if (tmp.length)
                           {
                           stats ~= tmp;
                           count += r.count;
                           }
                        }
                return stats;
        }

        /***********************************************************************

                Write the tree content to the provided IBuffer

        ***********************************************************************/

        private void emit (Parser.Node block, IBuffer buffer)
        {
                while (block)
                      {
                      buffer.append (block.content);
                      if (block.child)
                         {
                         buffer.append (block.bracket[0..1]);
                         emit (block.child, buffer);
                         buffer.append (block.bracket[1..2]);
                         }
                      block = block.next;
                      }
        }

        /***********************************************************************

                Support interception of class literals

        ***********************************************************************/

        protected void processLiteral (Parser.Node node)
        {
                node.type = node.Type.Literal;
        }

        /***********************************************************************

                Label the nodes with additional information. We also 
                identify specific keywords here; ones that will be
                required when reformatting the output.

        ***********************************************************************/

        private void label (Parser.Node block)
        {
                if ((block.label = classDecl (block.content)) !== null)
                   {
                   //printf ("class %.*s\n", block.label);
                   block.type = block.Type.Class;
                   block.info = new ClassInfo;
                   block.context = classDecl.match();
                   }
                else
                   if ((block.label = literalDecl (block.content)) !== null)
                        processLiteral (block);
                   else
                   if ((block.label = funcDecl (block.content)) !== null)
                        if (block.label in Keywords)
                           {
                           //printf ("statement %.*s\n", block.label);
                           block.type = block.Type.Statement;
                           }
                        else
                           {
                           //printf ("function %.*s\n", block.label);
                           block.type = block.Type.Function;
                           ++functions;
                           }
        }

        /***********************************************************************

                locate and rewrite constructors

        ***********************************************************************/

        private void constructor (Parser.Node block)
        {
                if (block.parent.type == block.Type.Class)
                    if (block.type == block.Type.Function)
                        if (block.label == block.parent.label)
                           {
                           Ctor ctor = new Ctor (block.label);
                           block.content = ctor (block.content);

                           // add this function to the ctor list
                           ClassAttrib ca = cast(ClassAttrib) cast(void *) block.parent.info;
                           assert (ca);
                           ca.ctors ~= block;
                           ++ctors;
                           }                
        }

        /***********************************************************************

                Handle package and import replacements

        ***********************************************************************/

        protected void preamble (Parser.Node block)
        {
                if (block.parent.type == block.Type.Root)
                    block.content = moduleName (block.content);
        }

        /***********************************************************************
        
                Perform general text replacements

        ***********************************************************************/

        protected void general (Parser.Node block)
        {
                if (block.parent.type == block.Type.Root)
                    block.content = interfaceDecl (block.content);

                block.content = instanceof(
                                literalStringEquals(
                                arrayCopy(toSwtString(string(
                                equalsNull(throws(renameBoolean(renameChar(
                                renameError(returnCast(uberCast(renameKeyword(block.content)))))))))))));
        }

        /***********************************************************************

                Handle package and import replacements

        ***********************************************************************/

        protected void cleanup (Parser.Node block)
        {
                   block.content = removeFinal(block.content);
        }

        /***********************************************************************

                Handle rewriting and movement of declaration initializers

        ***********************************************************************/

        private void decls (Parser.Node block)
        {
                Parser.Node     scope,
                                staticCtor;
                char[]          statics;
                char[]          instances;

                void declSpec (Parser.Node block)
                {
                        // convert array literals into template calls
                        block.content = newArrayInstance (block.content, block);

                        // find all the class declarations with 'new' in them.
                        // also find all class declaration with the template.
                        // don't eat declarations from inner classes!
                        if (block.parent is scope)
                           {
                           block.content = staticInitializer (block.content, statics);
                           block.content = instanceInitializer (block.content, instances);

                           // locate the static ctor also
                           char[] tmp = staticCtorDecl (block.content);
                           if (tmp.length && block.child && (block.child.type != block.Type.Comment))
                              {
                              // rewrite the static ctor with "static this()" 
                              // instead
                              block.content = tmp;

                              // if there are multiple static ctors, merge 
                              // them
                              if (staticCtor)
                                 {
                                 Parser.Node child = staticCtor.child;

                                 // should be at least one child!
                                 if (child)
                                    {
                                    // coalesce the previous static ctor with 
                                    // this one
                                    child.move (block);

                                    // drop prior static ctor and remove its 
                                    // declaration from the source
                                    staticCtor.child = null;
                                    staticCtor.content = staticCtorRemove (staticCtor.content);
                                    }
                                 }
                              staticCtor = block;
                              }
                           }
                }

                // if this node is a class, scan for both static and
                // instance declarations, and move them appropriately
                if (block.type == block.Type.Class)
                   {
                   scope = block;
                   if (scope.child)
                      {
                      // Is this class inside another? D inner classes do not 
                      // automatically host the parent pointer
                      if (scope.parent.type != scope.Type.Class)
                          currentParser.addClass (scope.label);
                      else
                         if (Text.indexOf (scope.context, "static") < 0)
                            {
                            currentParser.message ("non-static inner class '"~ scope.label ~"' may need manual connection to its parent");
                            currentParser.addClass (scope.parent.label~"."~scope.label);
                            }

                      // visit children to get class declarations
                      scope.child.visit (&declSpec);
                      scope.content = extends (scope.content);
                      }
                   else
                      // found a class decl in the wrong place!
                      currentParser.message ("found a bogus class decl '" ~ scope.label ~ "'");


                   // retrieve class info
                   ClassAttrib ca = cast(ClassAttrib) cast(void*) scope.info;
                   assert (ca);

                   // locate a ctor for instance-level assignments
                   if (instances)
                      {
                      Parser.Node ctor;
                      if (ca.ctors.length)
                         {
                         // could be a dependency issue here ...
                         if (ca.ctors.length > 1)
                             currentParser.message ("multiple ctors noted when adding ctor assignments (manual dependency check required)");

                         ctor = ca.ctors[0];
                         }
                      else
                         {
                         // create a default ctor
                         ctor = scope.addChild();
                         ctor.type = ctor.Type.Function;
                         ctor.content = "this ()";
                         ctor.addChild();
                         ++ctorsAdded;
                         }
                      // add the new assignments in front of other statements
                      ctor.child.content = instances ~ ctor.child.content;
                      }

                   // locate a static ctor for class-level assignments
                   if (statics)
                      {
                      if (! staticCtor)
                         {
                         // create a default static ctor
                         staticCtor = scope.addChild();
                         staticCtor.type = staticCtor.Type.Function;
                         staticCtor.content = "static this()";
                         staticCtor.addChild();
                         ++staticCtorsAdded;
                         }
                      // add the new assignments in front of other statements
                      staticCtor.child.content = statics ~ staticCtor.child.content;
                      }
                   }
        }
}


/*******************************************************************************

        Locate class declaration

*******************************************************************************/

class ClassDecl : Replace
{
        private Processor parent;

        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "classes scanned");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this (Processor parent)
        {
                super (r"^\s*((public|private|static|final|abstract)\s+)*(class|interface)\s+([\w_]+)", "m");
                this.parent = parent;
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content)
        {
                set (content);
                if (find)                              
                   {
                   ++count;
                   return groups.g4;
                   }
                return null;
        }
}

/*******************************************************************************

        Locate class declaration

*******************************************************************************/

class LiteralDecl : Replace
{
        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "class literals scanned");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r"new\s+(\w+)\s*\(\)\s*$");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content)
        {
                set (content);
                if (find)                              
                   {
                   ++count;
                   return groups.g1;
                   }
                return null;
        }
}

/*******************************************************************************

        Locate and patch interface declaration

*******************************************************************************/

class InterfaceDecl : Replace
{
        private Extends extends;

        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "interface declarations patched");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r"^\s*((public|private|final)\s+)*interface\s+([\w_]+)[^;]*", "m");
                extends = new Extends;
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content)
        {
                set (content);
                while (find)
                      {
                      ++count;
                      add ();
                      add (extends (match));
                      eat ();
                      }
                return combine;
        }
}

/*******************************************************************************

        Locate and patch static ctors

*******************************************************************************/

class StaticCtorDecl : Replace
{
        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "static ctors patched");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r"([^\w]static)\s*$");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content)
        {
                set (content);
                if (find)
                    with (groups())
                         {
                         ++count;
                         add ();
                         add (g1);
                         add (" this()");
                         eat ();
                         return combine;
                         }
                return null;
        }
}

/*******************************************************************************

        Remove unused static ctors (ones that were merged) 

*******************************************************************************/

class StaticCtorRemove : Replace
{
        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "static ctors merged");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r"[^\w]static this\(\)\s*$");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content)
        {
                set (content);
                if (find)
                   {
                   ++count;
                   add ();
                   eat ();
                   }
                return combine;
        }
}

/*******************************************************************************

        Locate function declarations. This also picks up certain keywords
        to be used when formatting the output, such as while() and if(). 

*******************************************************************************/

class FuncDecl : Replace
{
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
//                super (r"([\w_]+)\s*\(.*\)\s*$", "m");
//                super (r"([\w_]+)\s*\(.*\)\s*$", "m");
                super (r"([\w_]+)\s*\([^\)]*\)\s*$", "m");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content)
        {
                set (content);
                if (find)
                   {
                   return groups.g1;
                   }
                return null;
        }
}

/*******************************************************************************
        
        replace static int[] assignments with D version

*******************************************************************************/

class StaticIntArray : Replace
{
        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "static int[] initializers rewritten");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r"(static.*)= new int\[\] \{(.*)\};");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content)
        {
                set (content);
                while (find)
                       with (groups())
                            {
                            ++count;
                            add ();
                            add (g1);
                            add (" = [");
                            add (g2);
                            add ("];");
                            eat ();
                            }
                return combine;
        }
}

/*******************************************************************************
        
        replace new []{} assignments with a template instead

*******************************************************************************/

class NewArrayInstance : Replace
{
        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "array literal assignments rewritten");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r"\bnew\s+(\w+)\s*\[\]\s*$");
//                super (r"([^\w])new\s+(\w+)\s*\[\]\s*$");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content, Parser.Node block)
        {
                set (content);
                while (find)
                       with (groups())
                            {
                            ++count;
                            add ();
//                            add (g1);
                            add ("ConstructArray!(");
                            add (g1);
                            add (")(");
                            add (block.child.content);
                            add (")");
                            eat ();

                            // child node is now dead
                            block.child = null;
                            }
                return combine;
        }
}

/*******************************************************************************
        
        Change boolean to bool

*******************************************************************************/

class RenameBoolean : Replace
{
        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "boolean declarations patched");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r"\bboolean\b");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content)
        {
                set (content);
                while (find)
                      {
                      ++count;
                      add ();
                      add ("bool");
                      eat ();
                      }
                return combine;
        }
}

/*******************************************************************************
        
        eliminate 'final'

*******************************************************************************/

class RemoveFinal : Replace
{
        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "java keywords removed");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r"\b(final|native)\s+");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content)
        {
                set (content);
                while (find)
                      {
                      ++count;
                      add ();
                      eat ();
                      }
                return combine;
        }
}

/*******************************************************************************
        
        Change Error to RuntimeError

*******************************************************************************/

class RenameError : Replace
{
        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "Error references patched");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r"\bError\b");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content)
        {
                set (content);
                while (find)
                      {
                      ++count;
                      add ();
                      add ("RuntimeError");
                      eat ();
                      }
                return combine;
        }
}


/*******************************************************************************
        
        Process string literals

*******************************************************************************/

class String : Replace
{
        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "String literals patched");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r"\x22[^\x22]*\x22");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] text)
        {
                set (text);
                while (find)
                      {
                      add ();
                      char[] tmp = match;

                      char c;
                      int  end = end()-1;
                      int  start = start();
                      // ignore appended strings
                      while (--start >= 0)
                            {
                            c = content[start];
                            if (isspace(c))
                                continue;

                            // skip concatenations and arguments
//                            if (c == '+' || c == '(')
                            if (c == '+')
                               {
                               add (tmp);
                               goto next;
                               }
                            else
                               {
                               // skip '+=' also
//                               if (c == '=')
//                                   if (content[start-1] == '+')
//                                       continue;
                               break;                                              
                               }
                            }

                      // ignore "".equals() strings
                      while (++end < content.length)
                            {
                            c = content[end];
                            if (isspace(c))
                                continue;

                            // skip property reference (.equals .length)
                            if (c == '.')
                               {
                               add (tmp);
                               goto next;
                               }      
                            else
                               break;                                               
                          }

                      ++count;
                      if (tmp.length > 2)
                         {
                         add ("String.create(");
                         add (tmp);
                         add (")");
                         }
                      else
                         add ("String.Null");
next:
                      eat ();
                      }
                return combine;
        }
}

/*******************************************************************************

        replace extends with ':' 

*******************************************************************************/

class Extends : Replace
{
        bool                    colon;
        private Implements      implements;
          
        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "'extends' replaced");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r"extends");
                implements = new Implements (this);
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content)
        {
                colon = false;
                set (content);
                if (find)
                   {
                   ++count;
                   add ();
                   add (":");
                   colon = true;
                   eat ();
                   }
                return implements (combine);
        }
}

/*******************************************************************************

        replace implements with ':' or ','

*******************************************************************************/

class Implements : Replace
{
        private Extends parent;

        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "'implements' replaced");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this (Extends parent)
        {
                super (r"implements");
                this.parent = parent;
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content)
        {
                set (content);
                if (find)
                   {
                   ++count;
                   add ();
                   add (parent.colon ? "," : ":");
                   eat ();
                   }
                return combine;
        }
}

/*******************************************************************************

        Change ctor name to 'this'
         
*******************************************************************************/

class Ctor : Replace
{
        /***********************************************************************

        ***********************************************************************/

        this (char[] name)
        {
                super (name~r"\s*(\([^\)]*\)\s*$)");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] text)
        {
                set (text);
                while (find)
                      {
                      add ();
                      add ("this ");
                      add (groups.g1);
                      eat ();
                      }
                return combine;
        }
}

/*******************************************************************************

        Replace identity comparison 
         
*******************************************************************************/

class EqualsNull : Replace
{
        private NotEqualsNull   ne;

        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "'==' tests changed");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r"\s*==\s*");
                ne = new NotEqualsNull;
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content)
        {
                set (content);
                while (find)
                       with (groups())
                            {
                            ++count;
                            add ();
                            add (" is ");
                            eat ();
                            }
                return ne (combine);
        }
}

/*******************************************************************************

        Replace identity comparison 
         
*******************************************************************************/

class NotEqualsNull : Replace
{
        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "'!=' tests changed");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r"!=");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content)
        {
                set (content);
                while (find)
                       with (groups())
                            {
                            ++count;
                            add ();
                            add ("!==");
                            eat ();
                            }
                return combine;
        }
}

/*******************************************************************************

        Post processing package-scan to potentially add local package 
        imports (the ones Java does by default)

*******************************************************************************/

class PackagePass2 : Replace
{
        Processor parent;

        /***********************************************************************

        ***********************************************************************/

        this (Processor parent)
        {
                this.parent = parent;
                super (r"(module\s+[^;]+;)");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content, parent.ParsedFile file)
        {
                set (content);
                if (find)
                   {
                   add ();
                   add (match());
       
                   // import local packages?
                   char[] imports = parent.buildImport (file, file.packageMember);
                   if (imports.length)
                      {
                      add ("\n\n");
                      add (imports);
                      }
                   eat ();
                   }
                return combine;
        }

}


/*******************************************************************************

        Change 'package' into 'module' specification using the modified 
        filepath

*******************************************************************************/

class Package : Replace
{
        private Prefix          prefix;
        private Processor       parent;

        class Prefix : ImportPrefix
        {
                char[] stats (){return null;}
        }

        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "package names changed");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this (Processor parent)
        {
                super (r"package\s+([^;]+)");
                this.parent = parent;
                prefix = new Prefix;
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content)
        {
                static const char[] j2d = IMPORT_PREFIX ~ ".swt.util.Default";

                set (content);
                if (find)
                   {
                   ++count;
                   add ();
                   add ("\nmodule ");
                   char[] moduleName = prefix (parent.getModuleName);
                   add (moduleName);
                   add (";\nimport ");
                   add (j2d);

                   parent.PackageMember pm;

                   // get original package name
                   char[] packageName = groups().g1;

                   // have we seen this package before?
                   if (packageName in parent.packageMembers)
                       pm = parent.packageMembers [packageName];
                   else
                      {
                      // create new one if not
                      packageName = packageName.dup;
                      pm = new parent.PackageMember (packageName);
                      parent.packageMembers [packageName] = pm;
                      }
                   
                   // add this file to the package list
                   pm.addModule (parent.currentParser, moduleName.dup);
//                   printf ("'%.*s'\n", packageName);

                   // eat content
                   eat ();
                   }
                return combine;
        }
}

/*******************************************************************************

        Replace import statements 
         
*******************************************************************************/

class Import : Replace
{
        private ImportStar      star;
        private ImportJava      java;
        private ImportPrefix    prefix;
        private Processor       parent;

        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "imports made private");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this (Processor parent)
        {
                super (r"import\s*([^;]+);");

                prefix = new ImportPrefix;
                java = new ImportJava;
                star = new ImportStar;
                this.parent = parent;
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content, Processor.ParsedFile file)
        {
                set (content);
                while (find)
                       with (groups())
                            {
                            ++count;
                            add ();
//                            add (prefix(java(star(g1))));

                            int i = Text.indexOf (g1, '*');
                            if (i > 0)
                               {
                               parent.PackageMember pm;
                              
                               char[] packageName = g1[0..i-1];
                               if (packageName in parent.packageMembers)
                                  {
                                  pm = parent.packageMembers[packageName];
                                  char[] imports = parent.buildImport (file, pm);
                                  if (imports.length)
                                      add (imports);
                                  }
                               else
                                  {
                                  // printf ("rewrote unknown package %.*s\n", packageName);
                                  add ("private import ");
                                  add (prefix(java(star(g1))));
                                  add (";");
                                  }
                               }
                            else
                               {
                               add ("private import ");
                               add (prefix(java(g1)));
                               add (";");
                               }
                            eat ();
                            }
                return combine;
        }
}

/*******************************************************************************

        Change import prefix to something more appropriate

*******************************************************************************/

class ImportPrefix : Replace
{
        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "import prefixes replaced");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r".*org\.eclipse");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content)
        {
                set (content);
                while (find)
                      {
                      ++count;
                      add ();
                      add (IMPORT_PREFIX);
                      eat ();
                      }
                return combine;
        }
}

/*******************************************************************************

        Change import java.x to something more appropriate

*******************************************************************************/

class ImportJava : Replace
{
        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "import java.x replaced");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r"java\.");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content)
        {       
                static const char[] swt = IMPORT_PREFIX ~ ".swt.";

                set (content);
                while (find)
                      {
                      ++count;
                      add ();
                      add (swt);
                      eat ();
                      }
                return combine;
        }
}

/*******************************************************************************
        
        replace import * to import .all

*******************************************************************************/

class ImportStar : Replace
{
        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "import * replaced");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r"\.\*");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content)
        {
                set (content);
                while (find)
                      {
                      ++count;
                      add ();
                      add (".all");
                      eat ();
                      }
                return combine;
        }
}


/*******************************************************************************
        
        replace the entire import block!

*******************************************************************************/

class ImportReplace : Replace
{
        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "import block replaced");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r"(import\s+[^;]+;\s*)+", "m");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content, char[] replacement)
        {
                set (content);
                if (find)
                   {
                   ++count;
                   add ();
                   add (replacement);
                   eat ();
                   }
                return combine;
        }
}


/*******************************************************************************
        
        rewrite System.arraycopy() to include the type size

*******************************************************************************/

class ArrayCopy : Replace
{
        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "System.arraycopy instances patched");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r"System\.arraycopy\s*\((([^,]+).*)\)");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content)
        {
                set (content);
                while (find)
                       with (groups())
                            {
                            ++count;
                            add ();
                            add ("System.copy (");
                            add (g2);
                            add ("[0].sizeof, ");
                            add (g1);
                            add (")");
                            eat ();
                            }
                return combine;
        }
}

/*******************************************************************************
        
        rewrite toString() so it doesn't clash with Object.toString()

*******************************************************************************/

class ToSwtString : Replace
{
        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "toString() instances patched");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r"\btoString\b");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content)
        {
                set (content);
                while (find)
                      {
                      ++count;
                      add ();
                      add ("toSwtString");
                      eat ();
                      }
                return combine;
        }
}

/*******************************************************************************

        locate static initializers with new, rewrite and queue them up 
        for subsequent relocation

*******************************************************************************/

class StaticInitializer : Replace
{
        StaticInitializer2      si2; 
        StaticInitializer3      si3; 

        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "static 'new' initializers moved");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r"(static.*[\w\._\[\]]+\s+)([\w\._]+)\s*=\s*new\s*([^;]+)");
                si2 = new StaticInitializer2;
                si3 = new StaticInitializer3;
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content, inout char[] decls)
        {
                set (content);
                while (find)
                       with (groups())
                             if (g3.length)
                                {
                                ++count;
                                add ();
                                add (g1);
                                add (g2);
                                char[] tmp = "\t"~g2~" = new "~g3~"; // Moved\n";
                                decls ~= tmp;
                                eat ();
                                }
                return si3(si2(combine, decls), decls);
        }
}

/*******************************************************************************

        locate static initializers with ConstructArray() template, and 
        queue them up for subsequent relocation

*******************************************************************************/

class StaticInitializer2 : Replace
{
        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "static array literal assignments moved");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r"(static.*[\w\._\[\]]+\s+)([\w\._]+)(\s*= (ConstructArray|String\.create).*[^;]+)");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content, inout char[] decls)
        {
                set (content);
                while (find)
                       with (groups())
                             if (g3.length)
                                {
                                ++count;
                                add ();
                                add (g1);
                                add (g2);
                                char[] tmp = "\t"~g2~g3~"; // Moved\n";
                                decls ~= tmp;
                                eat ();
                                }
                return combine;
        }
}

/*******************************************************************************

        locate static initializers with "= foo ? bar : wumpus;" constructs 
        and queue them up for subsequent relocation

*******************************************************************************/

class StaticInitializer3 : Replace
{
        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "static ?: initializers moved");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r"(static.*[\w\.\_]+\s+)([\w_]+)(\s*=.*\?[^;]+)");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content, inout char[] decls)
        {
                set (content);
                while (find)
                       with (groups())
                             if (g3.length)
                                {
                                ++count;
                                add ();
                                add (g1);
                                add (g2);
                                char[] tmp = "\t"~g2~g3~"; // Moved\n";
                                decls ~= tmp;
                                eat ();
                                }
                return combine;
        }
}

/*******************************************************************************

        locate instance initializers with new, and queue them up for
        subsequent relocation

*******************************************************************************/

class InstanceInitializer : Replace
{
        private InstanceInitializer2    ii2;

        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "'new' initializers moved");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r"([\w\._\[\]]+)\s*=\s*new\s*([^;]+)");
                ii2 = new InstanceInitializer2;
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content, inout char[] decls)
        {
                set (content);
                while (find)
                       with (groups())
                             if (g2.length)
                                {
                                ++count;
                                add ();
                                add (g1);
                                char[] tmp = "\t"~g1~" = new "~g2~"; // Moved\n";
                                decls ~= tmp;
                                eat ();
                                }
                return ii2 (combine, decls);
        }
}

/*******************************************************************************

        locate instance initializers with ConstructArray(), and queue them 
        for subsequent relocation

*******************************************************************************/

class InstanceInitializer2 : Replace
{
        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "array literal assignments moved");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r"([\w\._]+)(\s*= (ConstructArray|String\.create).*[^;]+)");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content, inout char[] decls)
        {
                set (content);
                while (find)
                       with (groups())
                             if (g2.length)
                                {
                                ++count;
                                add ();
                                add (g1);
                                char[] tmp = "\t"~g1~g2~"; // Moved\n";
                                decls ~= tmp;
                                eat ();
                                }
                return combine;
        }
}

/*******************************************************************************

        Remove all 'throws'

*******************************************************************************/

class Throws : Replace
{
        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "'throws' removed");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r"throws.*");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content)
        {
                set (content);
                while (find)
                      {
                      ++count;
                      add ();
                      eat ();
                      }
                return combine;
        }
}


/*******************************************************************************

        Rewrite all "text".equals() with == instead 

*******************************************************************************/

class LiteralStringEquals : Replace
{
        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "\"text\".equals() rewritten");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (\x22 r"\.equals\s*\(([^\)]+)\)");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content)
        {
                set (content);
                while (find)
                       with (groups())
                            {
                            ++count;
                            add ();
                            add ("\" == ");
                            add (g1);
                            eat ();
                            }
                return combine;
        }
}


/*******************************************************************************

        Rewrite assignment casts

*******************************************************************************/

class AssignmentCast : Replace
{
        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "assignment casts patched");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r"=\s*(\([\s\w\._]+[\[\]]*\s*\)\s*[\w_\d]+)");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content)
        {
                return content;
                set (content);
                while (find)
                       with (groups())
                            {
                            ++count;
                            add ();
                            add ("= cast");
                            add (g1);
                            eat ();
                            }
                return combine;
        }
}


/*******************************************************************************

        Rewrite simple cast arguments & initializers 

*******************************************************************************/

class ArgumentCast : Replace
{
        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "argument/initializer casts patched");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r"(\([\s\w\._]+[\[\]]*\s*\)\s*[\w_\d]+)\s*([\n\r,}\)]+)");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content)
        {
                return content;
                set (content);
                while (find)
                       with (groups())
                            {
                            ++count;
                            add ();
                            add ("cast");
                            add (g1);
                            add (g2);
                            eat ();
                            }
                return combine;
        }
}


/*******************************************************************************

        Rewrite casts with (x)(((y))) format 

*******************************************************************************/

class ParensCast : Replace
{
        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "expression casts patched");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r"(\([\s\w\._]+[\[\]]*\)\s*\(+[^\)]+\)*)");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content)
        {
                return content;
                set (content);
                while (find)
                       with (groups())
                            {
                            ++count;
                            add ();
                            add ("cast");
                            add (g1);
                            eat ();
                            }
                return combine;
        }
}


/*******************************************************************************

        Rewrite casts of "return (x) y;" 

*******************************************************************************/

class ReturnCast : Replace
{
        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "other casts patched");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r"(return|throw)\s+(\([\w\._]+\))");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content)
        {
                set (content);
                while (find)
                       with (groups())
                            {
                            ++count;
                            add ();
                            add (g1);
                            add (" cast");
                            add (g2);
                            eat ();
                            }
                return combine;
        }
}


/*******************************************************************************

        Rewrite all other casts 

*******************************************************************************/

class UberCast : Replace
{
        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "expression casts patched");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r"\([\w\._]+(\s\[\])*\)");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] text)
        {
                set (text);
                while (find)
                      {
                      char c;

                      // scan for prior non-space char
                      for (int i=start(); --i >= 0;)
                          {
                          c = content [i];
                          if (! isspace (c))
                                break;
                          }

                      add ();

                      // don't cast if prefixed by identifier or by Template
                      if (! (isalnum(c) || c == '_' || c == '!'))
                         {
                         // scan for prior non-space char
                         for (int i=end(); i < content.length; ++i)
                             {
                             c = content [i];
                             if (! isspace (c))
                                   break;
                             }
                         // don't cast if followed by a '?'
                         if (c != '?')
                            {
                            ++count;
                            add ("cast");
                            }
                         }
                      add (match());
                      eat ();
                      }
                return combine;
        }
}


/*******************************************************************************

       

*******************************************************************************/

class Instanceof : Replace
{
        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "instanceof tests rewritten");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r"([\w\._]+\s*(\[\w+\])*)\s+instanceof\s+([\w\._]+)");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content)
        {
                set (content);
                while (find)
                       with (groups())
                            {
                            ++count;
                            add ();
                            add ("cast(");
                            add (g3);
                            add (") ");
                            add (g1);
                            eat ();
                            }
                return combine;
        }
}


/*******************************************************************************

        Prefix all keywords with _ to avoid D conflicts

*******************************************************************************/

class RenameKeyword : Replace
{
        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "keywords renamed");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                // removed struct & sizeof (May 25th)
                super (r"\b(version|bit|is|debug|in|out|length|union)\b");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content)
        {
                set (content);
                while (find)
                      {
                      ++count;
                      add ();
                      add ("_");
                      add (groups.g1);
                      eat ();
                      }
                return combine;
        }
}


/*******************************************************************************

        rename all char types to be Char, so they can be aliased

*******************************************************************************/

class RenameChar : Replace
{
        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "char instances patched");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r"\bchar\b");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content)
        {
                set (content);
                while (find)
                      {
                      ++count;
                      add ();
                      add ("Char");
                      eat ();
                      }
                return combine;
        }
}


/*******************************************************************************

        Rename all "union" instances to _union to avoid D conflicts

*******************************************************************************/

class RenameUnion : Replace
{
        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return format (count, "'union' instances renamed");
        }
      
        /***********************************************************************

        ***********************************************************************/

        this ()
        {
                super (r"\bunion\b");
        }

        /***********************************************************************

        ***********************************************************************/

        char[] opCall (char[] content)
        {
                set (content);
                while (find)
                      {
                      ++count;
                      add ();
                      add ("_union");
                      eat ();
                      }
                return combine;
        }
}


/*******************************************************************************

        Set of slices to return for group matching. See Replace.groups()

*******************************************************************************/

class Groups 
{
        public  char[]  g0,
                        g1,
                        g2,
                        g3,
                        g4,
                        g5,
                        g6,
                        g7,
                        g8,
                        g9;
}

/*******************************************************************************

        Wrapper around RegExp to handle regex-groups and text-chunks

*******************************************************************************/

class Replace : Groups
{       
        protected alias regmatch_t      Match;

        protected int                   count;
        protected RegExp                regex;
        protected char[]                content;
        protected char[][]              sections;
        protected static Replace[]      handlers;

        /***********************************************************************

        ***********************************************************************/

        this ()
        {
        }

        /***********************************************************************

        ***********************************************************************/

        this (char[] pattern, char[] attributes = null)
        {
                compile (pattern, attributes);

                // add this to our static list of handlers
                handlers ~= this;
        }

        /***********************************************************************

        ***********************************************************************/

        void compile (char[] pattern, char[] attributes = null)
        {
                regex = new RegExp (pattern, attributes);
        }

        /***********************************************************************
        
                Return statistics 

        ***********************************************************************/

        char[] stats ()
        {
                return null;
        }
      
        /***********************************************************************
        

        ***********************************************************************/

        static char[] format (int count, char[] msg)
        {
                char[8] tmp;
        
                if (count)
                    return Int.format(tmp, count) ~ "\t" ~ msg ~ "\n";
                return null;
        }
      
        /***********************************************************************

        ***********************************************************************/

        void set (char[] content)
        {
                this.content = content;
                sections.length = 0;
        }
        
        /***********************************************************************

        ***********************************************************************/

        char[] get ()
        {
                return content;
        }
        
        /***********************************************************************

        ***********************************************************************/

        bool find ()
        {
                if (regex.test (content))
                    return true;
                return false;
        }
        
        /***********************************************************************

        ***********************************************************************/

        void add (int group = 0)
        {
                sections ~= content [0 .. start(group)];
        }
        
        /***********************************************************************

        ***********************************************************************/

        void add (char[] other)
        {
                sections ~= other;
        }
        
        /***********************************************************************

        ***********************************************************************/

        void eat (int group = 0)
        {
                if (group < 0)
                    content = null;
                else
                   content = content [end(group) .. length];
        }
        
        /***********************************************************************

        ***********************************************************************/

        char[] combine ()
        {
                char[] combined;

                foreach (char[] section; sections)
                         combined ~= section;
                return combined ~ content;
        }
        
        /***********************************************************************

        ***********************************************************************/

        Groups groups ()
        {
                char[]* p = &g0;
                foreach (int i, Match m; regex.pmatch)
                        {
                        if (i > regex.re_nsub)
                            break;
                        int start = m.rm_so;
                        int end = m.rm_eo;
                        if (start >= 0 && end >= 0)
                            *p = content [start .. end];
                        else
                           *p == "";
                        ++p;
                        }
                return this;
        }
        
        /***********************************************************************

        ***********************************************************************/

        char[] match ()
        {
                int start = regex.pmatch[0].rm_so;
                int end = regex.pmatch[0].rm_eo;
                return content [start .. end];
        }
        
        /***********************************************************************

        ***********************************************************************/

        int start (int group = 0)
        {
                assert (group <= regex.re_nsub);
                return regex.pmatch[group].rm_so;
        }
        
        /***********************************************************************

        ***********************************************************************/

        int end (int group = 0)
        {
                assert (group <= regex.re_nsub);
                return regex.pmatch[group].rm_eo;
        }      
}

class S
{
        S opAdd (int s)
        {
                return this;
        }

        S opAdd (char[] s)
        {
                return this;
        }

        S opAddAssign (S s)
        {
                return this;
        }

        S opAddAssign (char[] s)
        {
                return this;
        }

        static final S create ()
        {
                return new S;
        }

        char[] opCast()
        {
         return "";
        }
}


/*******************************************************************************

        test zone ...

*******************************************************************************/

template ConstructArray(T)
{
        T[] ConstructArray (...)
        {
                T[] result = new T[_arguments.length];
                for (uint i=0; i < _arguments.length; ++i)
                    {
                    result[i] = *cast(T*) _argptr;
                    _argptr += T.sizeof;
                    }
                return result;
        }
}

void test(int[] dd)
{
        S s = new S;
        s = s + "" + 1;

        ConstructArray!(int)(1,2,3,4);
}

